# A rudimentary curses interface for the proxy
#   ** Can't handle window resizing **

from string import *
import signal, curses, sys, os

screen = None
ROWS = 25
COLUMNS = 80

# Since parts of the proxy still write to stdout, let's redirect stdout
# to a single line of the screen.  I can't seem to write to the last line
# of the screen, so write to the second to last ...
class CursesOutput:
    def __init__(self):
        self.line = ''
        
    def write(self, str):
        self.line = self.line + str
        i = rfind(self.line, '\n') # Print only up to the last newline
        if i >= 0: self.line = self.line[i+1:]
        if not self.line: return
        screen.addstr(ROWS-1, 0, (self.line+' '*COLUMNS)[:COLUMNS-1],
                      curses.A_STANDOUT)
        screen.move(ROWS-1, 0)
        screen.refresh()

class CursesError:
    def __init__(self):
        self.text = ''

    def write(self, str):
        self.text = self.text + str
        lines = split(self.text, '\n')
        for row in range(len(lines)):
            screen.addstr(row, 0, lines[row])
        screen.refresh()
        import time
        time.sleep(len(lines)/3)
        
_original_stdout = None
_original_stderr = None
_original_signal_handler = None

def _winch_handler(signum, frame):
    # Update the window size
    global ROWS, COLUMNS
    ROWS, COLUMNS = screen.getmaxyx()
    screen.clear()
    screen.refresh()
    
def init():
    if os.environ.get('TERM', '') == 'xterm':
        # Xterm titlebar:
        print '\033]2;Proxy\007'
        # Xterm icon name:
        print '\033]1;Proxy\007'
        # To set both use ]0;

    # Now initialize curses
    global screen
    screen=curses.initscr()

    # Remember the screen size
    global ROWS, COLUMNS
    ROWS, COLUMNS = screen.getmaxyx()

    # Whenever the window is resized, update the screen size
    global _original_signal_handler
    _original_signal_handler = signal.signal(signal.SIGWINCH, _winch_handler)
    
    # Turn off echoing of keys, and enter cbreak mode,
    # where no buffering is performed on keyboard input
    curses.noecho() ; curses.cbreak()

    # In keypad mode, escape sequences for special keys
    # (like the cursor keys) will be interpreted and
    # a special value like curses.KEY_LEFT will be returned
    screen.keypad(1)

    # Return -1 if getch() has no key to return
    screen.nodelay(1) 

    # Don't newline
    curses.nonl()
    
    # Clear the screen and prepare to write
    screen.clear()
    screen.refresh()

    global _original_stdout, _original_stderr
    _original_stdout = sys.stdout
    _original_stderr = sys.stderr
    sys.stdout = CursesOutput()
    sys.stderr = CursesError()

def exit():
    # Set everything back to normal
    screen.keypad(0)
    curses.echo() ; curses.nocbreak()
    curses.endwin()                 # Terminate curses

    # Restore the signal handler (not that there should be one ..)
    signal.signal(signal.SIGWINCH, _original_signal_handler)

    # Restore stdout
    sys.stdout = _original_stdout
    sys.stderr = _original_stderr

    # Call the previous exit function, if any
    _previous_exitfunc()

try: _previous_exitfunc = sys.exitfunc
except AttributeError: _previous_exitfunc = lambda:0 # empty function
sys.exitfunc = exit

def connection_to_key(conn):
    # The request number is unique and persists when the actual object
    # handling a connection changes, but if that doesn't exist, we use
    # the connection object as a fallback
    try: return conn.reqno
    except: return conn
    
def update_status_count(ready_count, total_count):
    screen.addstr(0, 0, (' Proxy | %2d/%2d connections | ' %
                         (ready_count, total_count) + ' '*COLUMNS)[:COLUMNS],
                  curses.A_REVERSE)

row_assignment = {}
last_row = 0

# Event handlers

def event_idle(connections):
    update_status_count(0, len(connections))
    
    # Go through the connections and erase the entries that are done
    known_keys = map(connection_to_key, connections)
    global last_row
    last_row = 0 # Reset last_row to the lowest used row on the screen
    for conn_key in row_assignment.keys():
        row = row_assignment[conn_key]
        if conn_key not in known_keys:
            del row_assignment[conn_key]
            screen.addstr(1+row, 0, ' '*COLUMNS)
        else:
            if row > last_row: last_row = row
            
    screen.move(0, COLUMNS-1)
    screen.refresh()
    
def event_status(ready_list, connections, conn):
    if ready_list is not None and connections is not None:
        event_idle(connections)
        update_status_count(len(ready_list), len(connections))
    conn_key = connection_to_key(conn)
    try:
        row = row_assignment[conn_key]
    except KeyError:
        # Find a new row for this thing
        global last_row
        rows_occupied = row_assignment.values()
        for i in range(ROWS-2):
            row = (i+last_row)%(ROWS-2)
            if row not in rows_occupied:
                row_assignment[conn_key] = row
                last_row = row
                break # Found a good row
        else:
            # Nowhere to put it!  Oh well.
            row = None
    if row != None:
        screen.addstr(1+row, 0, (conn.__repr__(COLUMNS)+' '*COLUMNS)[:COLUMNS])
        screen.move(1+row,COLUMNS-1)
    else:
        screen.move(0,COLUMNS-1)
    screen.refresh()
    
def event_ad_removed(url, tag):
    from proxy3_util import shorten_url
    h = 'Removed %s:' % tag
    print h, shorten_url(url, COLUMNS-1-len(h))
    
def event_name_lookup_begin(site):
    from proxy3_util import shorten_url
    print 'Lookup   %s ..' % shorten_url(site, COLUMNS-1-25)
    
def event_name_lookup_end(site, site_ip):
    from proxy3_util import shorten_url
    h =  ('Resolved %s -> %s' % (shorten_url(site, COLUMNS-1-25), site_ip))
    print h[:COLUMNS]

def event_unix_error():
    sys.stderr.write('\007') # beep
    
# Window methods:

# 'addch', 'addstr', 'attroff', 'attron', 'attrset', 'border', 'box',
# 'clear', 'clearok', 'clrtobot', 'clrtoeol', 'delch', 'deleteln',
# 'echochar', 'erase', 'getbegyx', 'getch', 'getmaxyx', 'getstr',
# 'getyx', 'hline', 'idlok', 'inch', 'insch', 'insertln', 'keypad',
# 'leaveok', 'move', 'mvwin', 'nodelay', 'nooutrefresh', 'notimeout',
# 'refresh', 'scroll', 'scrollok', 'setscrreg', 'standend', 'standout',
# 'subwin', 'touchline', 'touchwin', 'vline'
        
init()
