import array

from proxy3_filter import *
from proxy3_util import *
from time import time

class JavaClassFileFormat:
    """ Java Class File Format"""

    # Add all class substitutions here
    substs = {
	'java/awt/Frame': 'Safe$Frame',
	'Safe$Thread': 'Safe$Thread',
	'Safe$Socket':  'Safe$Socket',
	'NewCNN':  'NewCNN',
	'SafeApplet': 'SafeApplet',
	'AudioWindow': 'AudioWindow',
	'AudioClipClass': 'AudioClipClass',
	}

    def __init__(self, s, p, a, filter):
        self.s = s          # input string
	self.p = p          # current pointer to input string
	self.a = a
	self.filter = filter  # classify which class is used

    """ Primitive Functions such as read, write, transfer """
    def read(self,n):
	"""`Read' n bytes from the input string"""
	r = self.s[self.p:self.p+n]
	self.p = self.p+n
	# Pad the string with '\0' if needed; this is an ugly hack but
	# it helps us avoid errors!  :(
	return r + '\0' * (n-len(r))

    def write(self,s):
	"""`Write' a string to the output array"""
	self.a.fromstring(s)

    def transfer(self,n):
	"""Copy n bytes from the input to the output"""
	self.write(self.read(n))

    """ Functions dealing with CLASS FILE FORMAT """
    def read_magic(self):
        self.magic = self.read(4)
	return self.magic

    def write_magic(self):
        self.write(self.magic)
    
    def transfer_version(self):
        """ Trnasfer version number """
	self.transfer(4)

    def read_constant_count(self):
        """ Read Constant Count
	    Sometimes it does not go with the actual number of constant
	    pool entries. """
        str = self.read(2)
    	self.constant_count = ord(str[0])*256 + ord(str[1])  

	# ################# CHECK THIS
	# ## I've heard that the number of constants is one more than
	# ## the actual number of constants (!)
	self.constant_count = self.constant_count - 1

	return self.constant_count

    def write_constant_count(self):
    	self.write(chr((self.constant_count+1)/256)
		   +chr((self.constant_count+1)%256))

    def read_constant_pool(self):
	""" Read Constant Pool """

	pool = []
	pool.append ([''])
	skip = 0
	
	for i in range(1, self.constant_count+1):
	    if skip > 0:
	        skip = skip - 1
		continue
	    entry = []
	    flag = self.read(1)
	    entry.append (flag)
	    flag = ord(flag)
	    entry.append (flag)     # for debug

	    # Read one entry in the constant pool, and store it into a list

	    # ascii
	    if flag==1: 
	        strsize = self.read(2)
		size = ord(strsize[0])*256 + ord(strsize[1])
		str = self.read(size)

		# symbolic name susbtitution happens HERE
		if str in self.substs.keys():
		    # Change this string into another
		    str = self.substs[str]
		    strsize = chr(0) + chr(len(str))
		    ##print 'J        ==>   ',`len(str)`,`str`
		entry.append (strsize)
		entry.append (str)

	    # unicode
	    elif flag==2:
	        strsize = self.read(2)
		entry.append (strsize)
		size = ord(strsize[0])*256 + ord(strsize[1])
		str = self.read(2*size)
		entry.append (str)

	    # integer or float
	    elif flag==3 or flag==4:  
	        str = self.read(4)
		entry.append (str)

	    # CONSTANT_Long_info or CONSTANT_Double_info
	    # take up two entries.
	    elif flag==5 or flag==6: 
	        str = self.read(8)
		entry.append (str)
		skip = 1

	    # class, interface, or string
	    elif flag==7 or flag==8:  
	        str = self.read(2)
		entry.append (str)
		entry.append (ord(str[0])*256 + ord(str[1])) # for debug

	    # (interface) field, method, or abstract field and method
  	    elif flag==9 or flag==10 or flag==11 or flag==12:  
		str_name = self.read(2)
		entry.append (str_name)
		str_sig = self.read(2)
		entry.append (str_sig)
		# for debug
		entry.append (ord(str_name[0])*256 + ord(str_name[1]))
		entry.append (ord(str_sig[0])*256 + ord(str_sig[1]))

	    # undefined constant pool flag
	    else:
	        print '! ___   Unknown constant code, flag is',flag

	    # put this entry into constant pool list
	    pool.append (entry)
	    
	    # eight-byte constants, CONSTANT_Long_info and
	    # CONSTANT_Double_info take up two entries
	    if 5 <= flag <= 6:
	        pool.append (['\000', '0'])

	# ensure that we are keeping track of constant count well
	if len(pool)-1 != self.constant_count:
	    print '! ___   Inconsistent constant count ----------------'
	
	self.constant_pool = pool
	    
	return pool

    def modify_constant_pool(self):
        """ Modify Constant Pool 
	    Modify method references so that they can reference new
	    methods. We need to add symbolic names for class name, method
	    name(in general, though method name does not change), and
	    method type, and class reference entry for the new class """

        for i in range(1, self.constant_count):

	    """ It does NOT work to subsitute method invocation of 
	        interface methods. I can't figure out what difference
		makes our method-level substitution not work. """

	    # look for interface method reference entries in constant pool
	    if ord(self.constant_pool[i][0]) == 11:
	        # get symbolic names for the method referenced by this entry
	        class_name, method_name, method_type = self.get_method(i)

		# check if this method is
		#    AppletContext - showDocument - 
		if (class_name == 'java/applet/AppletContext' and
		    method_name == 'showDocument'):

		    method_type = (method_type[:1] +
			   'Ljava/applet/AppletContext;' + method_type[1:])

	            # replace it with a new mothod reference
		    self.replace_method(i, 'SafeApplet', 'showDocument',
					method_type)

	        # check if this method is 
		#    Applet - getAudioClip - 
		elif (class_name == 'java/applet/AudioClip' and
		      method_name == 'loop'):
		
		    method_type = (method_type[:1] +
			   'Ljava/applet/AudioClip;' + method_type[1:])

	            # replace it with a new method reference
		    self.replace_method(i,'SafeApplet','AudioLoop', 
					method_type)

	        # check if this method is 
		#    Applet - getAudioClip - 
		elif (class_name == 'java/applet/AudioClip' and
		      method_name == 'stop'):
		
		    method_type = (method_type[:1] +
			   'Ljava/applet/AudioClip;' + method_type[1:])

	            # replace it with a new method reference
		    self.replace_method(i,'SafeApplet','AudioStop', 
					method_type)




	    # look for method reference entries in constant pool
	    if ord(self.constant_pool[i][0]) == 10:
	        # get symbolic names for the method referenced by this entry
	        class_name, method_name, method_type = self.get_method(i)

	        # check if this method is 
		#    Applet - showStatus - (Ljava/lang/String)V
		if (class_name == 'java/applet/Applet' and
		      method_name == 'showStatus'):
		
	            # replace it with a new method reference
		    self.replace_method(i,'SafeApplet','showStatus', 
		    '(Ljava/applet/Applet;Ljava/lang/String;)V')

	        # check if this method is 
		#    Applet - getAudioClip - 
		elif (class_name == 'java/applet/Applet' and
		      method_name == 'getAudioClip'):
		
		    method_type = (method_type[:1] +
			   'Ljava/applet/Applet;' + method_type[1:])

	            # replace it with a new method reference
		    self.replace_method(i,'SafeApplet','getAudioClip', 
					method_type)



		# check if this method is Thread - setPriority - (I)V
		elif (class_name == 'java/lang/Thread' and
		    method_name == 'setPriority'):

		    # if so, replace this method reference with reference
		    # to a more safe method 
		    # Safe$Thread - setPriority - (Ljava/lang/Thread;I)V 
		    self.replace_method (i, 'Safe$Thread', 'setPriority',
					 '(Ljava/lang/Thread;I)V')

		# check if this method is java/net/Socket - <init>
		elif (class_name == 'java/net/Socket' and
		      method_name == '<init>'):

		    # make a method type for a new method
		    # whatever argument types are, append return type of
		    # object of class Socket 'Ljava/net/Socket;'
		    method_type = method_type[:-1] + 'Ljava/net/Socket;'

		    # replace method reference with reference to a new  method
		    # Safe$Socket - init - 'method_type'
 		    self.replace_method(i, 'Safe$Socket', 'init', method_type)


		if self.filter == 'CnnFilter':    

		    # add a method reference to a new method which is used
		    # for argument - level substitution
		    self.NewCNN_checkURL_entry = self.insert_method ('NewCNN', 
                           'checkUrl', '(Ljava/net/URL;)Ljava/net/URL;')

		    # check if this method is 
		    #    Graphics - drawString - (Ljava/lang/String;II)V 
		    if (class_name == 'java/awt/Graphics' and
			method_name == 'drawString'):
		
		        # replace it with a new method reference
			self.replace_method(i,'NewCNN','drawString', 
			'(Ljava/awt/Graphics;Ljava/lang/String;II)V')
			
		    # check if this method is 
		    #    Applet - showStatus - (Ljava/lang/String)V
		    elif (class_name == 'java/applet/Applet' and
			  method_name == 'showStatus'):
		
		        # replace it with a new method reference
			self.replace_method(i,'NewCNN','showStatus', 
			'(Ljava/applet/Applet;Ljava/lang/String;)V')



    def modify_bytecode (self):
        """ Modify Bytecode 
	    replace 'invokevirtual' with 'invokestatic' where we want to
	    substitute method reference. - Method reference modification
	    is done in constant pool<self.modify_constant_pool()> - """

        # Amit modified this function to use arrays instead of
	# lists of strings.  Arrays are MUCH faster!

	# Convert immutable string into mutable array of unsigned bytes
	bc = array.array('B', self.code)

	# This code would be faster if we simply searched for 182|185|187
	# in the original string :) -- Amit
	for i in xrange(len(self.code)):
	    # look for bytecode, 'invokevirtual' (=182)
	    if bc[i] == 182:
		# method reference entry in constant pool    
	        entry = 256*bc[i+1] + bc[i+2]
		try:
		    c_name, m_name, m_type = self.get_method(entry)
		except IndexError:
		    continue

		# check if this method invocation is for substituted
		# method reference entry

		# NewCNN - drawString
		# NewCNN - showStatus
		# Safe$Thread - setPriority
		if ( (c_name == 'Safe$Thread' and m_name == 'setPriority')
		     or (c_name == 'SafeApplet' and m_name == 'getAudioClip')
		     or (c_name == 'SafeApplet' and m_name == 'showStatus')):
		    # replace 'invokevirtual' with 'invokestatic'
		    bc[i] = 184       # invokestatic
		    print '-------------------------------------------------'
		    print c_name, m_name, m_type
		    print '-------------------------------------------------'

	    # look for bytecode, 'new' (=187)
	    elif bc[i] == 187:  # new
	        # class entry for 'new'
	        entry = 256*bc[i+1] + bc[i+2]
		try:
		    index = self.get_index(entry, 2)
		except IndexError:
		    continue
		class_name  = self.get_string(index)
		if class_name == 'java/net/Socket':
		    bc[i+3] = 87     # dup -> pop
		    j = 0
		    while bc[i+4+j] != 183:   # != invokespecial
		       j = j + 1
		    a = i + 4 + j
		    bc[a] = 184      # invokestatic
		    entry  = 256*bc[a+1] + bc[a+2]
		    print '--------------------------------------------------'
		    print self.get_method(entry)
		    print '--------------------------------------------------'

	    # look for bytecode, 'invokeinterface' (=185)
	    elif bc[i] == 185:  
		# method reference entry in constant pool    
	        entry = 256*bc[i+1] + bc[i+2]
		try:
		    c_name, m_name, m_type = self.get_method(entry)
		except IndexError:
		    continue

		# check if this method invocation is for substituted
		# method reference entry

		# NewCNN - showDocument
		if ( (c_name == 'SafeApplet') and 
		     (m_name == 'showDocument' or 
		      m_name=='AudioLoop' or
		      m_name=='AudioStop')):
		     
		    # replace 'invokeinterface' with 'invokestatic'
		    bc[i+4] = bc[i+2]
		    bc[i+3] = bc[i+1]
		    bc[i+2] = 184     # invokestatic
		    bc[i+1] = 0       # nop
		    bc[i]   = 0      # aload_0
		    print '+ + + + + + + + + + + + + + + + + + + + + + + +'
		    print c_name, m_name, m_type
		    print '+ + + + + + + + + + + + + + + + + + + + + + + +'

	# Convert back into an immutable string
	self.code = bc.tostring()


    def write_constant_pool(self):
	"""Write Constant Pool """
	for i in range(1, self.constant_count+1):
	    entry = self.constant_pool[i]
		
	    # skip this entry if flag is 0
	    if entry[0] == '\000': continue

	    # write flag
	    flag = entry[0]
	    self.write(flag)

	    # Write one entry in the constant pool
	    flag = ord(flag)
	    self.write(entry[2])
	    if (flag==1 or flag==2 or flag==9 or flag==10 or flag==11 
		or flag==12): 
		    self.write(entry[3])

    def transfer_access_flag(self):
        """ Transfer Access Flags With Printing Them Out """
	str = self.read(2)
	self.write(str)
	access_flag = ord(str[0])*256 + ord(str[1])
	access_flag = self.access_flags(access_flag)
	##print 'Access Flag:', `access_flag`
        
    def transfer_class_name(self):
        """ Transfer Classname With Printing It Out """
	str = self.read(2)
	self.write(str)
	this_class  = ord(str[0])*256 + ord(str[1])
	this_class  = self.get_index(this_class, 2)
	##print 'Class  Name:', self.get_string(this_class)

    def transfer_super_class(self):
        """ Transfer Superclass Name With Printing It Out """
	str = self.read(2)
	self.write(str)
	super_class = ord(str[0])*256 + ord(str[1])
	super_class  = self.get_index(super_class, 2)
	##print 'Super  Class:', self.get_string(super_class)

    def transfer_interface(self):
	""" Transfer Interface """    
	# Read the number of interfaces, a short int (2 bytes)
	str = self.read(2)
	self.write(str)
	interface_count = ord(str[0])*256 + ord(str[1])  
	##print 'Interface Count:', `interface_count`

	# Read interfaces
	for i in range(interface_count):
	    str = self.read(2)
	    self.write(str)
	    interface = ord(str[0])*256 + ord(str[1])
	    interface = self.get_index(interface, 2)
	    ##print '      Interface:', self.get_string(interface)
	    
    def transfer_fields(self):
        """ Transfer Fields """
	# Read the number of field, a short int (2 bytes)
	str = self.read(2)
	self.write(str)
	field_count = ord(str[0])*256 + ord(str[1])  
	##print 'Field Count:', `field_count`

	# Read fields (4 items, 2 bytes each)
	for i in range(field_count):
	    # get access flag, name, type, and attribute count
	    access_flag, name, type, count = self.transfer_header('.field ')
	    # Read Attribute
	    for j in range(count):
	        str = self.read(8)
		self.write(str)
		# Attribute Name
		namei = ord(str[0])*256 + ord(str[1])
		name = self.get_string(namei)
		# Attribute Length
		length = ord(str[4])*256 + ord(str[5])
		# index to constant pool
		index = ord(str[6])*256 + ord(str[7])
		##print '            ', name, `self.constant_pool[index][1]`
	    
    def transfer_methods(self):
        """ Transfer Methods """
	# Read the number of method, a short int (2 bytes)
	str = self.read(2)
	self.write(str)
	method_count = ord(str[0])*256 + ord(str[1])  
	##print 'Method Count:', `method_count`

	# Read methods (4 items, 2 bytes each)
	for i in range(method_count):
	    # get access flag, name, type, and attribute count
	    access_flag, name, type,count = self.transfer_header('.method ')
	    # just for debug
	    self.method_name = name
	    # Read Attribute
	    for j in range(count):
	        self.attribute()


    def attribute(self):
        # attribute name_index (2 bytes)
	ni = self.read(2)
	self.write(ni)
	name_index = ord(ni[0])*256 + ord(ni[1])
	attr_name = self.get_string(name_index)

	# attribute length (4 bytes)
	str= self.read(4)
	attr_length = ( ord(str[0])*256*256*256 + ord(str[3]) +
			ord(str[1])*256*256 + ord(str[2])*256 )

	# We are going to modify this attribute if this is code attribute for
	# a method, ""CNN_Headlines.mouseDown""
	# We are going to add some extra bytes of bytecodes into this
	# attribute. (Currently 22 bytes) 
	if (self.filter == 'CnnFilter' and self.method_name == 'mouseDown'
	    and attr_name == 'Code'):
            l = attr_length + 22
            str = chr(l/256/256/256) + chr(l/256/256) + chr(l/256) + chr(l%256)
	    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	    
	self.write(str)
	#print '           attribute name', `attr_name`, '-', `attr_length`, 'bytes'

	# classify attribute according to its name
	if attr_name == 'Code':
	    a_count = self.code_attribute()
	    #sys.stdout.write('     ')
	    for i in range(a_count):
		self.attribute()

	elif ( attr_name == 'SourceFile'
	       or attr_name == 'ConstantValue'
	       or attr_name == 'Exceptions'
	       or attr_name == 'LineNumberTable'
	       or attr_name == 'LocalVariableTable' ):
	    self.transfer(attr_length)
	#else:
	    ##print 'Unknown attribute name:', attr_name

    def modify_CNN_mouseDown(self, code_length):
        """ Modify the codes of method, CNN_Headlines.mouseDown, so that
	    we can substitute interface method invocation.
	    What we are trying to do here is to substitute arguments of 
	    method called 'showDocument' to load our intended documents. """

        code = []
	skip = 0
        for i in range(code_length):
	    if skip > 0:
	        skip = skip - 1
		continue
		
	    # look for invoke_interface (=185)
	    if ord(self.code[i]) == 185:
	        # method reference entry in constant pool    
		entry = 256*ord(self.code[i+1]) + ord(self.code[i+2])
		try:
		    c_name, m_name, m_type = self.get_method(entry)
	        except IndexError:
		    continue

		# make sure that this interface method invocation is for
		# method, showDocument
		if (c_name == 'java/applet/AppletContext' and
		    m_name == 'showDocument'):
		    
		    # branch to the end of code
		    code.append(chr(167))       # goto
		    offset = len(self.code) - i
		    code.append(chr(offset/256))
		    code.append(chr(offset%256))

		    # interface method invocation needs five steps of
		    # bytecodes where as any other needs only three steps. 
		    # so we should clear out the rest 2 bytes since the
		    # first three bytes are used for 'goto' bytecode 
		    code.append(chr(0))
		    code.append(chr(0))
		    skip = 4

		    # add method invocation to our method to the end of code
		    self.code.append(chr(184))       # invoke static
		    self.code.append(chr(self.NewCNN_checkURL_entry/256))
		    self.code.append(chr(self.NewCNN_checkURL_entry%256))

		    # move mothod invocation to showDocument to the end
		    # again, interface method invocation needs five steps
		    self.code.append(self.code[i])
		    self.code.append(self.code[i+1])
		    self.code.append(self.code[i+2])
		    self.code.append(self.code[i+3])
		    self.code.append(self.code[i+4])
		    
		    # branch to the place where was right after
		    # showDocument was originally invoked
		    self.code.append(chr(167))  # goto
		    offset = 65535 - offset - 3
		    self.code.append(chr(offset/256))
		    self.code.append(chr(offset%256))
	    else:
	        code.append (self.code[i])

	return code

    def code_attribute(self):
        # max_stack, max_locals (2 byte each)
	self.transfer(4)
	# code_length (4 bytes)
	str = self.read(4)
	code_length = ( ord(str[0])*256*256*256 + ord(str[3]) +
			ord(str[1])*256*256 + ord(str[2])*256 )

        # code[code_length]
	self.code = self.read(code_length)

	# pass the length of code to the browser
	self.write(str)

	# modify bytecode, then write it to the browser
	self.modify_bytecode()
	self.write(self.code)

	# exception table length (2 bytes)
	el = self.read(2)
	self.write(el)
	exception_length = ord(el[0])*256 + ord(el[1])  
	# exception table
	for k in range(exception_length):
	    self.transfer(8)
	# attribute count (2 bytes)
	ac = self.read(2)
	self.write(ac)
	attr_count = ord(ac[0])*256 + ord(ac[1])  
	    
	# return attribute_count
	return attr_count


    def transfer_header(self, s):
        """ parse info per field and method
	    access flags, name, type, attribute count """
        str = self.read(8)
	self.write(str)
	# Access Flags
	access_flagi = ord(str[0])*256 + ord(str[1])
	access_flag = self.access_flags(access_flagi)
	# Name
	namei = ord(str[2])*256 + ord(str[3])
	name = self.get_string(namei)
	# Type
	typei = ord(str[4])*256 + ord(str[5])
	type = self.get_string(typei)
	# Attribute Count
	count = ord(str[6])*256 + ord(str[7])
	##print '    ', s, name, type, `access_flag`
	return access_flag, name, type, count


    def access_flags(self, access_flag):
        """ Parse Access Flags 
	    return a list of access flags """

        access_list = []
	if access_flag & 0x0001:
		access_list.append('PUBLIC')
	if access_flag & 0x0002:
		access_list.append('PRIVATE')
	if access_flag & 0x0004:
		access_list.append('PROTECTED')
	if access_flag & 0x0008:
		access_list.append('STATIC')
	if access_flag & 0x0010:
		access_list.append('FINAL')
	if access_flag & 0x0020:
		access_list.append('SYNCHRONIZED')
	if access_flag & 0x0040:
		access_list.append('VOLATILE')
	if access_flag & 0x0080:
		access_list.append('TRANSIENT')
	if access_flag & 0x0100:
		access_list.append('NATIVE')
	if access_flag & 0x0200:
		access_list.append('INTERFACE')
	if access_flag & 0x0400:
		access_list.append('ABSTRACT')

	return access_list
    
    def get_index (self, i, j):
        index = ( 256 * ord(self.constant_pool[i][j][0]) +
		  ord(self.constant_pool[i][j][1]) )
    	return index

    def get_string (self, index):
        return self.constant_pool[index][3]
    
    def get_p (self):
        return self.p

    def get_method (self, index):
        # check if this entry is method reference entry
	"""
	if (self.constant_pool[index][1] < 9 or
	    self.constant_pool[index][1] > 11):
	    ##print '* * *', `index`, `self.constant_pool[index]`
	    raise IndexError
        """

	# CONSTANT_Class
        class_index = self.get_index(index, 2)
	class_name_index = self.get_index (class_index, 2)
	class_name = self.get_string(class_name_index)
	# CONSTANT_NameAndType
	nt_index = self.get_index(index, 3)
	method_name_index = self.get_index(nt_index, 2)
	method_name = self.get_string(method_name_index)
	method_type_index = self.get_index(nt_index, 3)
	method_type = self.get_string(method_type_index)
	# return method info
	return class_name, method_name, method_type

    def add_constant_entries(self, class_name, method_name, method_type):
        """ Add constant entries for a new method """
        # put ascii entry for class name 
	strsize = chr(0) + chr(len(class_name))
	self.constant_pool.append([chr(1), 1, strsize, class_name])

	# put ascii entry for method name
	strsize = chr(0) + chr(len(method_name))
	self.constant_pool.append([chr(1), 1, strsize, method_name])
	method_name_entry = self.constant_count + 2 

	# put ascii entry for method type
	strsize = chr(0) + chr(len(method_type))
	self.constant_pool.append([chr(1), 1, strsize, method_type])
	method_type_entry = self.constant_count + 3 

        # put class entry for class
	class_index = self.constant_count + 1 
	str_index = chr(class_index/256) + chr(class_index%256)
	self.constant_pool.append([chr(7), 7, str_index])
	class_entry = self.constant_count + 4 

	# adjust constant count (increased by 4)
	self.constant_count = self.constant_count + 4

	# return indice
	return class_entry, method_name_entry, method_type_entry

    def insert_class(self, class_name):
        # put ascii entry for class name 
	strsize = chr(0) + chr(len(class_name))
	self.constant_pool.append([chr(1), 1, strsize, class_name])

        # put class entry for class
	class_index = self.constant_count + 1 
	str_index = chr(class_index/256) + chr(class_index%256)
	self.constant_pool.append([chr(7), 7, str_index])
	class_entry = self.constant_count + 2 

	# adjust constant count (increased by 4)
	self.constant_count = self.constant_count + 2

	# return indice
	return class_entry


    def insert_method(self, class_name, method_name, method_type):
    
        # add constant entries for class name, method name, method type
        (class_entry, method_name_entry, method_type_entry) = (
	    self.add_constant_entries (class_name, method_name, method_type))

	class_entry = chr(class_entry/256) + chr(class_entry%256)
	name_entry = chr(method_name_entry/256) + chr(method_name_entry%256)
	type_entry = chr(method_type_entry/256) + chr(method_type_entry%256)

	# add constant entry for method name and type
	self.constant_pool.append( (chr(12), 12, name_entry, type_entry) )
	NT_entry = self.constant_count + 1
	NT_entry = chr(NT_entry/256) + chr(NT_entry%256)
	
	# add constant entry for method reference
	self.constant_pool.append( (chr(10), 10, class_entry, NT_entry) )

	self.constant_count = self.constant_count + 2

	return self.constant_count   # method reference entry

    def replace_method(self, entry, class_name, method_name, method_type):
        """ Substitue method reference to an unsafe method with
	    method reference to a more safe method """

        (class_entry, method_name_entry, method_type_entry) = (
            self.add_constant_entries (class_name, method_name, method_type))
	
	""" Replace method entry so that the new method is referenced """
	# substitute class reference
	self.constant_pool[entry][2] = ( 
		chr(class_entry/256) + chr(class_entry%256) ) 
	NameType_index = self.get_index(entry, 3)
	# substitute method name index
	self.constant_pool[NameType_index][2] = ( 
		chr(method_name_entry/256) + chr(method_name_entry%256))
	# substitute method type index
	self.constant_pool[NameType_index][3] = ( 
		chr(method_type_entry/256) + chr(method_type_entry%256))

	# make sure that entry tag is method reference
	self.constant_pool[entry][0] = chr(10)
	self.constant_pool[entry][1] = 10



    def print_bytecode(self):
	length = len(self.code)
	##print 'length', `length`
	constant_pool = self.constant_pool
	bytecodes = self.bytecodes
        skip = 0
	for i in range(length):
	    # skip this time... 
	    # this is already interpreted as an operand
	    if skip > 0: 
	        skip = skip - 1
		continue
	    opcode = ord(self.code[i])
	    #print `opcode`
	    # --------------
            # method invocatinos -  invokevirtual, invokespecial, etc.
	    if 180<=opcode<= 185: # heh, a <= b <= c doesn't work in java
		    index = 256*ord(self.code[i+1]) + ord(self.code[i+2])
		    method = self.get_method(index)
		    stdout.write("       " + `i` + "  ")
		    stdout.write(bytecodes[opcode]+" - ")
		    stdout.write(method[0]+"\n")
		    stdout.write("                         - ")
 		    stdout.write(method[1] +
				 "  " +
				 method[2] + "\n")
		    skip = 2
            # -------------
	    # new
	    elif opcode == 187: 
		    class_index = 256*ord(self.code[i+1]) + ord(self.code[i+2])
		    name_index = self.get_index (class_index, 2)
		    class_name = self.get_string(name_index)
		    #index = 256*codes[i+1] + codes[i+2]
		    stdout.write("       " + `i` + "  ")
		    stdout.write(bytecodes[opcode]+" - ")
		    stdout.write(class_name + "\n")
		    skip = 2
		    
	    # -------------
	    # bytecodes for branch - if, goto, etc.
	    elif 153 <= opcode <= 167 or 198<= opcode <= 199:
		    offset = 256*ord(self.code[i+1]) + ord(self.code[i+2])
		    stdout.write("       " + `i` + "  ")
		    stdout.write(bytecodes[opcode] + 
				 " - goto ")
		    ##print `offset`
		    skip = 2

	    # -------------
	    # goto wide index
	    elif opcode == 200:
		    offset = (256*256*256*ord(self.code[i+1]) + 
			      256*256*ord(self.code[i+2]) +
			      256*ord(self.code[i+1]) + ord(self.code[i+2]))
		    stdout.write("       " + `i` + "  ")
		    stdout.write(bytecodes[opcode] + 
				 " - ")
		    ##print `offset`
		    skip = 4
		    
	    elif 19 <= opcode <= 20:
		    skip = 2
	    elif (0 <= opcode <=15 or 159<=opcode<=177 or
		  181<= opcode <=187 or 38<=opcode<=45 or 
		  87<=opcode <= 89 or opcode==50 or opcode==76):
	            stdout.write("       " + `i` + "  ")
		    stdout.write(bytecodes[opcode]+"\n")
	    elif opcode == 192:
		    stdout.write("       " + `i` + "  ")
		    stdout.write(bytecodes[opcode]+"\n")
	    else: 
		    stdout.write("       " + `i` + "  ")
		    ##print `opcode`

    bytecodes = {0:'nop', 1:'aconst_null', 2:'iconst_m1', 
		 3:'iconst_0', 4:'iconst_1',  5:'iconst_2', 
		 6:'iconst_3', 7:'iconst_4',  8:'iconst_5', 
		 9:'lconst_0',10:'iconst_1', 11:'fconst_0', 
		 12:'fconst_1',13:'fconst_2', 14:'dconst_0', 
		 15:'dconst_1', 
		 21:'iload', 22:'lload', 23:'fload',
		 24:'dload', 25:'aload',
		 32:'lload_2',
		 38:'dload_0', 39:'dload_1', 40:'dload_2', 
		 41:'dload_3', 
		 42:'aload_0', 43:'aload_1', 44:'aload_2', 
		 45:'aload_3', 
		 50:'aaload',
		 76:'astore_1',
		 87:'pop', 88:'pop2', 89:'dup',
		 153:'ifeq', 154:'ifne', 155:'ifge', 
		 156:'ifge', 157:'1fgt', 158:'ifle',
		 159:'if_icmpeq', 160:'if_icmpne',
		 161:'if_icmplt', 162:'if_icmpge',
		 163:'if_icmpgt', 164:'if_icmple',
		 165:'if_acmpeq', 166:'if_acmpne',
		 167:'goto', 168:'jsr', 169:'ret',
		 170:'tableswitch', 171:'lookupswitch',
		 172:'ireturn', 173:'lreturn', 174:'freturn',
		 175:'dreturn', 176:'areturn', 177:'return',
		 180:'getfield', 181:'putfield',
		 182:'invokevirtual', 183:'invokespecial',
		 184:'invokestatic', 185:'invokeinterface',
		 186:'xxxunusedxxx',
		 187:'new', 
		 192:'checkcast', 193:'instanceof',
		 198:'ifnull', 199:'ifnonnull', 200:'goto_w',
		 224:'checkcast_quick'}
    


class JavaClassSubstituter:
    """Transform Java bytecode in the string 's', by performing
    each of the substitutions in substs, a mapping from str->str,
    and then return the modified string"""

    # Add all class substitutions here
    substs = {
	'java/awt/Frame': 'Safe$Frame',
	'Safe$Thread': 'Safe$Thread',
	'Safe$Socket':  'Safe$Socket',
	'NewCNN':  'NewCNN',
	'SafeApplet': 'SafeApplet',
	'AudioWindow': 'AudioWindow',
	'AudioClipClass': 'AudioClipClass',
	}

    def __init__(self, s, filter):
	self.s = s  # input string
	self.a = array.array('c')  # output array
	self.p = 0  # position in the input
	
	self.filter = filter  # current used filter
	self.jcff = JavaClassFileFormat(s, self.p, self.a, filter)

	before_time = time()

	# Transfer the magic number
	magic = self.jcff.read_magic()
	if magic != '\xca\xfe\xba\xbe': 
	    # Not a Java file - write the rest of it as is
	    self.jcff.write(self.s)
	    return

	# magic
	self.jcff.write_magic()

	# version
	self.jcff.transfer_version()

	# Read the number of constants, a short int (2 bytes)
	self.constant_count = self.jcff.read_constant_count()
	
	#print 'J       (Processing',len(s),'byte Java class with ',self.constant_count,'constants)'

	# constant pool
	self.constant_pool = self.jcff.read_constant_pool()

	after_time = time()
	time_elapsed = after_time - before_time
	print '|       Class-level modification took %2.5fs' % time_elapsed

	# Modify constant pool
	self.jcff.modify_constant_pool()
	self.jcff.write_constant_count()
	self.jcff.write_constant_pool()

	# Handle info about access flag(2), class(2), and super class(2)
	# self.jcff.transfer(6)
	self.jcff.transfer_access_flag()
	self.jcff.transfer_class_name()
	self.jcff.transfer_super_class()

	# Interface
	self.jcff.transfer_interface()
	# Fields
	self.jcff.transfer_fields()

	# Methods
	self.jcff.transfer_methods()
	
	# Transfer the rest of the bytecode
	self.p = self.jcff.get_p()
	self.write(self.s[self.p:])

	after2_time = time()
	time_elapsed = after2_time - after_time
	print '|       Method-level modification took %2.5fs' % time_elapsed

	    
    def read(self,n):
	"""`Read' n bytes from the input string"""
	r = self.s[self.p:self.p+n]
	self.p = self.p+n
	# Pad the string with '\0' if needed; this is an ugly hack but
	# it helps us avoid errors!  :(
	return r + '\0' * (n-len(r))

    def write(self,s):
	"""`Write' a string to the output array"""
	self.a.fromstring(s)
	#self.fo.write(s)

    def transfer(self,n):
	"""Copy n bytes from the input to the output"""
	self.write(self.read(n))

    def output(self):
	"""Turn the output array back into a string and return it"""
	return self.a.tostring()

class TestFilter(BufferAllFilter):
    def filter(self, s):
	print '----------------------------------------------------------'
	print '   T E S T - F I L T E R'
	print '----------------------------------------------------------'

	j = JavaClassSubstituter(s, 'TestFilter')
	return j.output()

class JavaClassFilter(BufferAllFilter):
    def filter(self, s):
	j = JavaClassSubstituter(s, 'JavaClassFilter')
	return j.output()

""" currently NOT used routine """
	    
"""
	    # look for bytecode, 'invokeinterface' (=185)
	    if ord(self.code[i]) == 185:  
		# method reference entry in constant pool    
	        entry = 256*ord(self.code[i+1]) + ord(self.code[i+2])
		try:
		    c_name, m_name, m_type = self.get_method(entry)
		except IndexError:
		    continue

		# check if this method invocation is for substituted
		# method reference entry

		# NewCNN - showDocument
		if ( (c_name == 'SafeApplet') and 
		     (m_name == 'showDocument' or 
		      m_name=='AudioLoop' or
		      m_name=='AudioStop')):
		     
		    # replace 'invokeinterface' with 'invokestatic'
		    self.code[i+4] = self.code[i+2]
		    self.code[i+3] = self.code[i+1]
		    self.code[i+2] = chr(184)     # invokestatic
		    self.code[i+1] = chr(0)       # nop
		    self.code[i]   = chr(0)      # aload_0
		    ##print '+ + + + + + + + + + + + + + + + + + + + + + + +'
		    ##print c_name, m_name, m_type
		    ##print '+ + + + + + + + + + + + + + + + + + + + + + + +'
"""

def event_process_request(obj, site, document, request, headers):
    # Redirection of safe classes
    end_name = split(document,'/')[-1]
    if end_name[-6:] == '.class' and  find(site,'~ishin/proxy') < 0:
        # This is a Java class, so let's determine whether it needs
        # to be redirected.
        if split(end_name[:-6], '/')[-1] in JavaClassSubstituter.substs.values():
            # This class is one of our inserted classes, so it should
            # be loaded from a different place
            site = 'www-cs-students.stanford.edu'
            document = '/~ishin/proxy/' + split(document,'/')[-1]
            print '|  ===> Redirecting to',site+document
            return site, document
    
def event_process_reply(obj, site, document, reply, headers):
    # modify Content-length if this document looks like Java class
    if document[-6:] == ".class":
	# This is a really bad way of doing it ....
	try:
	    length = atoi(headers['Content-length'])
	    length = length + 200
	    set_header(headers, 'Content-length', `length`)
	except KeyError:
	    try:
		length = atoi(headers['Content-Length'])
		length = length + 200
		set_header(headers, 'Content-Length', `length`)
	    except KeyError: pass


from proxy3_filter import register_filter
register_filter('*/test.class','*',TestFilter)
register_filter('*/*.class','*',JavaClassFilter)
register_filter('*/*.class','application/java-vm',None)
register_filter('*','application/java-vm', JavaClassFilter)
register_filter('*/~ishin/proxy/PUI.class','*',None)
register_filter('*/~ishin/proxy/PUIWindow.class','*',None)
register_filter('*/~ishin/proxy/SocketClient.class','*',None)

# Add all the exceptions to the substitution table - ew
for x in JavaClassSubstituter.substs.values():
    register_filter('*/'+x+'.class','*',None)
