# Implement a basic search engine over surfing history

#   Indexing: every time we see a page, we parse it, split the non-HTML
#   text into words, and associate the page's URL with each of the words.

#   Searching: we build up a list of URLs from each of the words entered
#   and then we rank them based on word frequency.  Like Google, we only
#   implement AND search.

from string import *
from proxy3_util import *
from proxy3_filter import *
import cgi, REX

INDEX = {} # Map words to URLs
TITLES = {} # Map URLs to titles

class IndexWords(BufferSomeFilter):
    # We don't actually modify any of the content, so we don't really
    # build up a buffer that's sent to the user agent.  However, we're
    # going to parse and index the whole thing later, so we need to
    # save the page.

    BUFFER_LEN = 0 # We don't really want buffering
    page_content = None

    def filter(self, s):
        # Save everything that goes by
        if not self.page_content: self.page_content = []
        self.page_content.append(s)

        return s
    
    def close(self):
        BufferSomeFilter.close(self)
        
        url = self.args['url']
        page_content = join(self.page_content, '')
        # Extract the page title if we can
        m = re.search(r'(?is)<title>(.*)</title>', page_content)
        if m: TITLES[url] = m.group(1)

        # Now it's time to parse and index the entire document
        parsed = REX.ShallowParse(page_content)
        in_script = 0
        in_style = 0
        in_head = 0
        for text_or_tag in parsed:
            maybe_tag = lower(text_or_tag)
            if maybe_tag[:7] == '<script': in_script = 1
            if maybe_tag[:8] == '</script': in_script = 0
            if maybe_tag[:6] == '<style': in_style = 1
            if maybe_tag[:7] == '</style': in_style = 0
            if maybe_tag[:5] == '<head': in_head = 1
            if maybe_tag[:6] == '</head': in_head = 0

            #TODO: actually we don't ignore the <head> section
            
            if not in_script and not in_style and text_or_tag[:1] != '<':
                # '<' means it's a tag, and we ignore them,
                # and we also ignore everything in <style>, <script>, <head>
                words = re.split(r'[^\w]+', text_or_tag)
                for word in words:
                    word = lower(word)
                    if not word: continue
                    if not INDEX.has_key(word):
                        INDEX[word] = {} # Make a new entry URL->count
                    dict = INDEX[word]
                    dict[url] = dict.get(url, 0) + 1

def handle_search(obj, q):
    # Display the form first
    q = strip(q)
    output = ['<html><title>Proxy Search over History: %s</title>'
              '<body><form action="/search" method=get>'
              'Search history: <input type=text value="%s" name=q>'
              '<input type=submit value="Go"></form>'
              % (cgi.escape(q), cgi.escape(q))]
    # Display results, if q is non empty
    if q:
        # Assume AND queries, like Google
        urls = [] # (Score, URL) pairs
        not_found = []
        for word in re.split(r'[^\w]+', q):
            word = strip(lower(word))
            if not word:
                pass # Ignore this; it was punctuation or spaces
            elif not INDEX.has_key(word):
                not_found.append(word)
            else:
                results = INDEX[word].items() # (URL, count) pairs
                max_count = 0 # Simple rescaling relative to max count
                for _, count in results:
                    max_count = max(max_count, count)
                urls.append([])
                for url, count in results:
                    # Save the rescaled results
                    urls[-1].append( (1.0*count/max_count, url) )

        # Now combine the results by taking only URLs in all lists
        if not_found:
            output.append('<p>These terms could not be found:  ')
            output.append(join(not_found, ', '))
            output.append('</p>')
        elif not urls:
            output.append('<p>No terms to search</p>')
        else:
            # Map URLs to (score, count)
            scoring = {}
            for group in urls:
                for score, url in group:
                    old = scoring.get(url, (0, 0))
                    scoring[url] = (old[0] + score, old[0]+1)
            # Now prune anything that wasn't in all the lists
            for url, (_, count) in scoring.items():
                if count != len(urls):
                    del scoring[url]
                    
            if not scoring:
                # No results
                output.append('<p>No pages found.</p>')
            else:
                results = scoring.items()
                results.sort(lambda x,y: cmp(x[1], y[1]))
                results.reverse() # Highest first
                for url, score in results:
                    output.append('<br> <a href="http://%s">%s</a>' %
                                  (url, TITLES.get(url, url)))
                    
    output.append('</body></html>')
    html_output(obj.outfile, join(output, ''))
    return []

def event_process_request(obj, site, document, request, headers):
    if site=='_proxy' and document[:7]=='/search':
        # Now look for q=
        args = cgi.parse_qs(document[8:])
        q = args.get('q', [''])[0]
        return handle_search(obj, q)

register_filter('*', 'text/html', IndexWords)
