#!/usr/bin/python

# HTTP proxy - Amit Patel, amitp@cs.stanford.edu
#  Thu 14 Aug 1997 - Initial design using event-based programming style
#                    ** SEE THIS PAGE TO READ ABOUT WHY THIS DESIGN WORKS **
#                    ** http://www.nightmare.com/medusa/async_sockets.html **
#                    In an odd twist of fate, two and a half years
#                    later, the author of Medusa (Sam Rushing) is now
#                    my coworker.  :-)
#  Fri 15 Aug 1997 - Implemented single connection proxy.
#  Sun 17 Aug 1997 - Implemented multiconnection proxy.
#                  - Extensive testing
#  Mon 18 Aug 1997 - Added special URLs:
#                    http://_proxy/location          - retreives host/port
#                    http://_proxy/start/*           - redirects to ishin
#                  - More testing
#  Fri 22 Aug 1997 - Added special URL:
#                    http://_proxy/start/            - redirects to /ui/
#                    http://_proxy/ui/*              - redirects to ishin
#                  - Also, the /start/ URL will append the hostname to the
#                    URL so that applets will be granted permission to
#                    communicate with that host.
#                  - Any request for a local document will be redirected
#                    to http://theory.stanford.edu/
#  Mon 25 Aug 1997 - Fixed bug that caused sockets to block when partial
#                    data was received from a web server.
#  Tue 26 Aug 1997 - Added test code for non-GET requests from UI applet
#  Thu 28 Aug 1997 - Investigated the POST form method, which requires
#                    changes to the proxy data transfer system.
#  Fri 29 Aug 1997 - Simplified the sending of HTML responses by 
#                    writing html_output function, which adds proper
#                    headers automatically.
#                  - Simplified the handling of blocked sites by
#                    using glob patterns instead of explicit if/else code.
#                  - Added the HTTPRequest handler, which takes over
#                    handling from Request if it's GET/POST
#                  - Fixed a bug with form handling:  the connection
#                    must be *blocking* in order for the HTTPRequest
#                    class to work.  It can't receive partial data
#                    as it's processing the headers.
#  Mon  8 Sep 1997 - Split CopyFile into several handlers:  CopyData,
#                    HTTPReply, DataReader, and BufferedWriter.
#                  - HTTPReply reads in HTTP reply headers and writes
#                    them out.  The headers are needed because important
#                    information, such as content-type, is used to
#                    dispatch on the file type.
#                  - DataReader just reads in data from a connection.
#                  - CopyData writes the data to an output connection.
#                  - BufferedWriter waits until all data is read before
#                    writing the data out.  It also runs the data through
#                    a series of filters.
#                  - Analyzed the design of these new classes and determined
#                    that poor performance would result
#                  - Redesigned data copying classes
#  Tue  9 Sep 1997 - Added an Options class that encapsulates all user
#                    options
#                  - Rewrote Excite Ad filter to fit into the BufferedWriter
#                    & DataReader framework
#                  - Moved blocked site functions and data into a separate
#                    class
#  Wed 10 Sep 1997 - Added the JavaClassFilter to perform class substitutions
#                    on Java classes, and tested it on some of Sun's applets
#  Thu 11 Sep 1997 - Added Safe$Class redirection to redirect requests for
#                    certain classes to a different site
#                  - Added timing code to see which handlers are taking a
#                    long time
#  Wed 18 Sep 1997 - Fixed bug in Java class substitution:  if the class 
#                    being loaded is a substituted class, no further
#                    substitutions should be made
#  Wed 28 Oct 1998 - Added GeoCities filter
#                  - Abstracted out HTML filtering using patterns
#                  - Added removal of IFRAME ads
#  Fri 30 Oct 1998 - Updated Excite filter to reflect Excite's redesign
#                  - Updated ad filters
#                  - Fixed regexp for ad filtering to handle linebreaks
#                  - Added removal of LAYER ads
#                  - Added removal of SCRIPT ads
#                  - Added more information about which URLs are being loaded

# PROXY VERSION 3 - Started Fri 30 Oct 1998
#  
#  Fri 30 Oct 1998 - Requires Python 1.5 (updated httplib, new regexps)
#                  - Changed Options class into options dictionary
#                    (This is nicer when we want to refer to options w/strings)
#                  - Added an Event handler system
#                    Various modules can register themselves as being
#                    interested in certain events.
#                  - Moved many parts of the program into separate modules,
#                    and those modules register themselves to receive events
#                  - Added a cookie reporting system that reports when
#                    a page sends a cookie to the browser
#                  - Changed HTTPReply to know about the user's address
#                  - HTTPReply now looks at Content-Length and passes
#                    this information on to CopyData
#                  - CopyData reads at most Content-Length bytes
#                  - CopyData reports on the progress of the transfer
#                  - Experimenting with using signals for timeouts on DNS
#                  - Added a filter for Tripod sites
#                  - The GeoCities & Tripod filters now chain->GenericAdFilter
#                  - Ad removal (GenericAdFilter) is more robust, finds more
#  Sat 31 Oct 1998 - Added a special URL http://_proxy/headers, which
#                    tells you about the headers your browser is sending.
#                  - Moved the event system into a separate module
#                  - Fixed a bug in the ad filtering; it was missing all
#                    patterns that didn't have http:// in them.
#                  - Simplified the connection handlers by having them
#                    copy all the fields from their parents instead of
#                    explicitly passing them.
#                  - Fixed a MAJOR bug with the way socket errors were
#                    being handled.  It turns out that Unix can return
#                    an error code even when there's no error.  The
#                    app is responsible for distinguishing non-error 
#                    error codes and real-error error codes.  The proxy
#                    works much better now, especially on long pages
#                    being loaded over a modem.
#                  - If the operating system supports Unix-like signals,
#                    the proxy will install a signal handler to time out
#                    from socket connections that take a long time to open.
#                    (Unfortunately, it doesn't automatically retry, so I
#                    made the timeout pretty long.)
#                  - Improved the error message that is shown when a
#                    connection fails.  Now there is a Retry link to retry.
#                  - Improved reporting of errors
#                  - Redid the POST form handling and tested it; it seems
#                    to work but it hasn't been tested extensively.
#  Sun  1 Nov 1998 - Added statistics gathering module
#                  - Added support for removing FORM ads
#  Fri  6 Nov 1998 - Moved filter table out into a separate module
#                    HTML & Java filters are now registered only when loaded
#                  - Added a special URL http://_proxy/filters to tell you
#                    what filters are being used
#                  - Moved basic HTTP core into proxy3_base, proxy3_web
#                  - Changed the interface for filters from string -> string
#                    to string * {keyword args} -> string
#  Mon  9 Nov 1998 - Adopted Jun Zhai's streamable filter model.
#                    The filters are no longer functions, but instead are
#                    objects that can filter as they read data.  (This is
#                    a big plus for HTML filters.)  Instead of a list of
#                    filters, each filter maintains a pointer to the output
#                    stream.  Any filter that can work on small pieces of
#                    data (and can also tolerate re-filtering the same data
#                    more than once) can inherit from BufferSomeFilter.
#                    Any filter that requires all the data at once should
#                    instead inherit from BufferAllFilter.  Filters can also
#                    derive from the Filter base class
#                  - Rewrote the HTML filters to derive from BufferSomeFilter
#                  - Rewrote the Java filters to derive from BufferAllFilter
#  Mon 16 Nov 1998 - Added more patterns to the HTML ad filtering
#  Sat 21 Nov 1998 - Fixed bugs in the java filtering (there was a
#                    recursive substitution.. crashed browsers)
#  Fri  4 Dec 1998 - Added a "proxy auto-config" system, using 
#                    http://<machine>:3333/auto .  It's not much use
#                    since manual configuration is just as easy.  Also,
#                    it doesn't work to exclude https, ftp, etc. yet.
#  Sun  6 Dec 1998 - Fixed proxy to use \r\n instead of \n, since the
#                    HTTP spec says we must use \r\n, at least for headers.
#                    Only a few servers seem to mind \n alone.
#  Thu 10 Dec 1998 - Added support to remove <A HREF="ad site">...</A> links
#                  - Removed Excite ad filtering, since the generic ad
#                    filter can remove all of Excite's ads
#                  - Fixed a bug in ad filtering of relative URLs
#  Sat 12 Dec 1998 - Added http://_proxy/blocklist to show all the patterns
#                    being blocked.
#                  - Fixed a bug in ad filtering, to handle mixed case
#                    tags.  For example, <A> .. </a> now is handled properly.
#  Wed 16 Dec 1998 - Added a 'ping' method to handler objects.  The core
#                    proxy can ask a handler to test its connections to 
#                    make sure they are not stale.  The handler returns
#                    a list of handlers to include in the new connection list.
#                  - Changed the HTML filter to increase its window size
#                    if an ad appears to be very long

# PROXY VERSION 3.1 - Sun 20 Dec 1998
#  
#  Sun 20 Dec 1998 - Added a configuration file ('config') that should store
#                    all configuration information
#                  - Moved the blocked URL list to the config file
#                  - Moved the list of optional modules to the config file
#                  - Moved the list of proxy flags to the config file
#                  - If host lookup fails and there are no periods in the
#                    hostname, the proxy will prepend www. and append .com
#                  - Host lookup is performed four times, each with timeout
#                  - Implemented a simple DNS lookup cache that keeps each
#                    name for 5 minutes.
#  Fri 25 Dec 1998 - Added a timeout for stale connections
#                  - Moved most of the printing to stdio out to module stdio
#                  - Moved the statistics gathering out to module stats
#                  - Cookie reporting is now a separate module, and can
#                    report both incoming and outgoing cookies
#                  - Enhanced the event module to allow for automatic
#                    scanning of modules (using introspection) and 
#                    registering of events, if the functions are named
#                    event_a_b_c for event a-b-c.
#                  - Changed the optional module loading system to 
#                    automatically register events in loaded modules.
#                    Now optional event modules do not have to register
#                    their handlers manually, as long as they follow
#                    the naming conventions.
#                  - Renamed all the optional modules to mod_*
#                  - Added a module for reporting load progress
#  Sat 26 Dec 1998 - Fixed a bug with filter streaming.  The filters were
#                    sending data to the browser instead of to the next filter
#                  - Implemented reuse of sockets (Keep-Alive extension to
#                    HTTP/1.0; also part of HTTP/1.1) in proxy3_socket.py
#                    Unfortunately it doesn't work well on some sites.
#                  - Fixed a bug in data transfering.  It seems that 
#                    feeding a number to socket.recv that exactly matches
#                    the number of bytes in the buffer .. can cause some
#                    strange errors to occur elsewhere.  (!?!)
#                  - Changed the sockets to use binary mode instead of 
#                    text mode.  I don't know if it matters in Unix,
#                    but for portability it's safer using binary mode.
#  Mon 28 Dec 1998 - Experimented with persistent connections, then gave up
#                    after much hacking.
#                  - Added a config file option to specify HTTP headers
#                    NOT to display (for the mod_headers module).
#                  - Renamed the config file to proxy.conf because
#                    that's more standard.
#  Fri  1 Jan 1999 - Fixed bug with reading tiny documents.  The data was
#                    in the same packet as the headers, and select() wasn't
#                    telling us that something was available because it was
#                    already in the buffer.
#                  - Added error handling for EINTR, a Unix error telling
#                    us some resource was busy and we have to try again.
#  Sat  9 Jan 1999 - Changed the output format for connections.  Instead of
#                    '@' for ready, '.' for not ready, it simply shows how
#                    many are ready out of the total number of connections.
#                  - Added a priority system for handlers.  Listeners have
#                    high priority, HTML gets medium-high priority, images
#                    get low priority, and other connections get medium
#                    priority.  This may help on overloaded modem connections.
#  Fri 15 Jan 1999 - Added an option to disable the HTTP header rewriting.
#  Thu 21 Jan 1999 - Fixed a major bug in the data reader.  The bug only
#                    appeared with fast connections, so that's why I
#                    never found it before.
#                  - Changed the HTML filtering to use one giant regexp
#                    instead of a hundred separate ones.  The regexp
#                    compiler might be able to optimize it better.
#  Sun 24 Jan 1999 - Fixed a bug in the text replacement filter that filtered
#                    out only the first pattern found.
#                  - Added a timeout to the connection code.  This closes
#                    the (hopefully) last place the proxy will freeze for
#                    long periods of time.
#                  - Added a new HTTPConnect handler that is used to
#                    connect to a web site.  The advantage of this is
#                    that we can continue doing things while the connection
#                    is being established.  Since establishing a connection
#                    is slow, we don't want to block on it.
#  Tue 26 Jan 1999 - Added error handling to the new HTTPConnect class
#                  - Changed the error report to the user to use larger
#                    fonts.
#  Sat 30 Jan 1999 - Added a change-dir command to change to the directory
#                    where the proxy was loaded.  This allows us to find
#                    data files.  Thanks to Richard Thrapp for the patch.
#                  - Added an error message if the proxy is already running.
#                    Thanks to Richard Thrapp for the idea.
#                  - Fixed a bug in URLs that use colons but aren't
#                    specifying a port number.
#                  - Added recognition for non-http URLs, but no support yet.
#                  - Uncaught exceptions during handlers are now
#                    printed to stdout but the proxy continues to run.
#  Fri  7 May 1999 - Changed the proxy autoconfig to use the proxy if
#                    it is there, or use a direct connection if it can't
#                    find the proxy.
#                  - Added an HTML filter to remove Microsoft-specific
#                    quote characters, and replace them with standard ASCII
#  Fri 14 May 1999 - Changed the proxy autoconfig to use the proxy only
#                    for http:// requests (not ftp:// or https://).
#                  - Changed the proxy autoconfig to disable the proxy for
#                    local pages.
#  Sat 29 May 1999 - Disabled the Pragma: no-cache attempt by servers to
#                    prevent caching.  This is questionable in general but
#                    it seems to help a lot, especially with slashdot.
#                  - Finally fixed the proxy auto-config to do:
#                       1.'yahoo' goes to the proxy
#                       2.'yahoo/' is considered local (no proxy)
#                       3.'yahoo.stanford.edu' is considered local (no proxy)
#                       4.'_proxy/' goes to the proxy
#  Wed  9 Jun 1999 - Added a Slashdot filter that removes some extra space
#                  - Added a PopupAdFilter to filter out popup ads (but not
#                    all popups) and also <BODY onUnload=...> evil hacks;
#                    the special case GeoCities and Tripod code is no longer
#                    needed, and has been removed.
#                  - Added error checking in the HTTP Listener
#  Sun 20 Jun 1999 - Added support for mime types with options (X/Y; Z)
#                  - HTML comments are preserved in the ad filter, since
#                    the filter only understands HTML, and HTML-like text
#                    may appear inside JavaScript.
#  Thu 24 Jun 1999 - Added special support for link rewriting, to replace
#                    */fwd/*/http://* with http://*
#  Sat 26 Jun 1999 - Added more Slashdot filtering as an experiment in 
#                    altering GIF files to match color changes on the page.

# PROXY VERSION 3.2 - Sun  4 Jul 1999
#  
#  Sun  4 Jul 1999 - Added a curses interface (mod_curses)
#  Mon  5 Jul 1999 - Changed the formatting of some of the __repr__ text
#                    and also made it support arbitrary line widths
#  Wed 14 Jul 1999 - Added a special www.google.com/frame URL that will
#                    set up Google in framed mode (so that the left frame
#                    shows search results and the right frame shows the pages)
#  Thu 15 Jul 1999 - Fixed a bug with handling of errors when port != 80
#  Sat 17 Jul 1999 - Removed proxy3_socket, a failed experiment to have
#                    persistent connections.
#                  - Changed the autoconfig to use the proxy whenever a
#                    a full hostname is given.
#                  - Fixed some bugs with the _proxy/* URLs
#                  - If an event handler returns something than None,
#                    further handlers will not be run, and the return
#                    value will be returned from broadcast_event.
#                  - Cleaned up the redirection -- it's now modular, with
#                    the process-request event allowing (1) hacks on the
#                    headers; (2) returning a pair (site, document) for
#                    redirection, or (3) returning a list of connection
#                    objects, to signal that the request was handled in
#                    some way (if it's [], the current connection is closed).
#                  - Removed Insik's Java/UI experiment from the proxy core
#                  - Made the _proxy/filters output a table
#                  - Removed the filter dependence on configuration variables,
#                    because with loadable modules you can simply choose
#                    the modules you want instead of both choosing modules
#                    and setting the corresponding variables.
#                  - The special URL proxy:* now maps to http://_proxy/*
#                    except that if * ends in /, it won't in the remapping.
#  Fri 23 Jul 1999 - Separated the Deja.com and Slashdot colorization filters
#                    into separate modules.
#  Sat 14 Aug 1999 - When choosing filters, we use the original URL instead
#                    of the URL after internal redirection
#                  - Made many modifications to the Google filter to
#                    make the /ie frame use popup windows with descriptions
#  Tue 17 Aug 1999 - The HTTP objects now keep track of the referer, so
#                    that we can perform filtering based on referer.
#                  - Highlighting of search keywords: if you perform a
#                    (Google) search, then click on a result, the proxy
#                    will attempt to highlight keywords on the resulting page
#                    The problem is that this mod_highlight module doesn't
#                    parse HTML, so it can alter URLs and HTML tags, and
#                    many pages end up broken.
#  Thu  9 Sep 1999 - Fixed bugs in the google module for /left, /frame, /table
#  Sat 11 Sep 1999 - The curses module now uses all the lines for display
#  Sat  2 Oct 1999 - Modified BufferSomeFilter to allow the filter() function
#                    to specify which part of the string should not be kept
#                    in the buffer.  This enables strings to be filtered
#                    only once.
#                  - Added mod_prefetchdns, a module that attempts to
#                    scan the links on the page and perform DNS lookup
#                    on the hosts mentioned in those links.  This should
#                    speed up access when you actually click on the link.
#  Sun  3 Oct 1999 - Fixed a bug in the <EMBED> filtering.
#                  - Tracebacks are now logged to /tmp/proxy.traceback
#                  - Fixed the mod_block module, which never worked.
#                  - The command line parsing now handles options (name=value)
#                  - The proxy port is now configurable in the configuration
#                    file (and also at the command line)
#                  - A "parent" proxy can be set up by setting the 'proxy'
#                    configuration option.
#                  - Added many configuration options for fine-tuning
#                    the ad filtering.
#                  - Added mod_gtk, a module that brings up a Gtk interface
#                    to the proxy.  It runs the Gtk window in a separate
#                    thread, and currently allows you to change some
#                    of the configuration options.  It will also show
#                    the status of the proxy and its connections.
#  Sat  9 Oct 1999 - Added proxy3_rewrite, a module that can rewrite
#                    "clickthrough" URLs based on the [rewrite] section
#                    of the configuration file.
#  Tue  9 Nov 1999 - Fixed handling of http://proxy/http://proxy requests
#                    wrt filtering
#  Fri 14 Jan 2000 - Added mod_search, a simple search engine over the
#                    pages the proxy has seen.  Often times you may think,
#                    "I've seen this page recently" and want to get back
#                    to it... mod_search can help you do that.
#  Sat 15 Jan 2000 - Added mod_cache, a simple in-memory document cache.
#                    Ideally, cached HTML would be sent to the browser
#                    with a "document.window=..." or other sort of meta-
#                    refresh that would load a more recent copy later.
#  Thu 20 Jan 2000 - Fixed a bug with POSTs; sometimes we need a newline
#                    terminator
#  Thu 20 Jan 2000 - The proxy now passes both the client (browser) headers
#                    and the server headers to filters.
#  Thu 20 Jan 2000 - Rewrote mod_highlight to work modularly, and also
#                    to handle parsing HTML.  It only highlights keywords
#                    in text, and not in tags, scripts, or style sheets.
#  Thu 27 Jan 2000 - Added mod_nocookie, which allows you to block cookies
#                    from selected sites (listed in the [nocookies] section
#                    of the configuration file)
#                  - Added a hack -- if the page is a 404, instead of seeing
#                    the server's 404 content, you get redirected to Google's
#                    cached copy of the page (assuming it has one).
#  Thu 20 Apr 2000 - Merged in Oleg Broytmann's <phd@phd.russ.ru>
#                    cookie configuration: instead of allowing all cookies
#                    except certain sites, allow no cookies except certain
#                    sites.  Set the mode with accept-cookie-default conf.
#  Sat 11 Nov 2000 - Fixes for Python 2.0
#                  - Support non-quoted command line arguments, a=x is a="x"
#  Sun 12 Nov 2000 - In proxy3_dns, check not only for signal but SIGALRM

# TO DO:
#  When you click on a document from a search engine, it should
#    highlight keywords in the resulting document (!!).  We would
#    need to choose filters not only by site+document+mimetype but
#    also Referer.
#  Add better support for rewriting headers BOTH ways
#  Add better control over changing Referer and User-Agent
#  Support HTTP/1.1 on the server side.  See
#      http://www.colorado.edu/infs/jcb/sinewave/network/http1.1/
#    for reasons.
#  Split the configuration file into several, where the main one
#    can #include other files.
#  Remove <SCRIPT>, <EMBED>, <BLINK>, <OBJECT>, on selected sites
#  Support redirection in the policy file, using http code 302
#  JavaScript editing: replace navigator.appName with something else,
#    to confuse ad servers.  Replace document.write with something that
#    checks for advertisements at run time (!!).
#  Support <BASE HREF> when completing ad URLs.

# TO DO LATER:
#  Total time elapsed for reading a document
#  Make the writing to the browser use asynchronous sockets
#  Use the time elapsed to close stale connections, or instead, use time IDLE
#  Configure which side of a low bandwidth link we're on.  If we're
#    on the server side, we want to continue reading a document and
#    dump it into cache, even if the browser isn't ready; if we're
#    on the client side, we want to abort as soon as possible to save
#    bandwidth
#  Asynchronous DNS would be nice
#    However, we now have asynchronous connections, and a DNS cache...
#  Add a module for sending messages to a log file
#  Figure out how to handle the Content-Length header.  Java classes
#    get longer.  HTML files get shorter.  If we really want to make
#    the header accurate, we can't do incremental filtering.

# TO DO, LONGER TERM:
#  Make per-user preferences (per-machine, actually) 
#    Use cookies to track user if he moves from machine to machine
#  Configure preferences through HTML forms
#  Add predicates to the policy file
#  Add a cache? (per user?)
#  Add a (per-user) history of visited pages
#  Automatically reload modules/policy-files that have been updated
#    The Jigsaw proxy parses the policy file and writes out a binary
#    version, so that you don't have to reparse every time.
#  Limit the number of connections, and implement a priority queue
#    that prefers images near the TOP of an HTML file, or also scan
#    IMG tag and load small images first (?)
#  Scan GIF files and see if they're animated; if so, remove animation bit
#  If a page has ads, we could tell the proxy we want to filter ads.
#    Then the proxy gives the browser a hacked version of the page, with
#    all the existing links removed, and new links added for every image.
#    These new links would go to the proxy and tell it which ads should 
#    be removed.

# BUGS/LIMITATIONS:
#  - The ad filtering doesn't take <BASE HREF> into account.
#  - If the HTML constructs an advertisement using Javascript, the filter
#    won't have removed it.  However, it will block loading of the image.
#  - Doesn't handle PUT and DELETE requests.
#  - *********** WARNING ***********
#    If the HTTP headers being sent to the server are large, the resulting
#    reply from the server will be mangled (by the server or by the proxy
#    I do not know).  This may have something to do with non-blocking when
#    sending headers.
#  - Modifying HTML and Java bytecode may violate copyrights.  I think
#    it's allowed for personal use, but you cannot redistribute the
#    HTML (i.e., run this proxy server for a group of users).

# Notes:
#  - See http://www.nightmare.com/medusa/async_sockets.html for information
#    about the event-based model used in this Proxy.
#  - A socket is closed if upon recv(1,socket.MSG_PEEK), it 
#    raises (_,'Bad file number')
#  - The CONNECT method should be routed to a different request handler
#  - Binary (and some text) data includes a 'Content-length' header.
#    However, dynamically generated text cannot have it.  HTTP/1.1
#    allows for 'chunked' data, which embeds the content length into
#    the content.
#  - The User-Agent header can include multiple words, each of the form
#    <name> [/<version>].  The proxy could append "MURI-Python-Proxy/??"
#    but it's just going to confuse browser detection on the server side,
#    and some pages may not work properly.  Also a privacy risk.
#  - Should we be using 's.shutdown(1)' to close just one side of a
#    socket?  It seems to work except on old Google.com (Medusa
#    server).  Google now uses a different server.

from string import *
from time import time
from sys import stdout
import sys, re, select, socket, os

# Change to the directory that contains the proxy, so that we
# can find data files -- thanks to Richard Thrapp
if sys.path[0]: os.chdir(sys.path[0])

######################################################################

def proxy():
    # At first, spawn two listeners, one for HTTP connections and the
    # other for UI connections.
    original_connections = []
    for listen_class, listen_port in proxy3_options.listeners:
        original_connections.append( listen_class(listen_port) )
    
    print 'Starting proxy on',proxy3_base.HOST,'; press Ctrl-C to stop.'
    connections = original_connections[:] # make a copy of the list
    timeout = 0.1
    while 1:
        readable = filter(lambda x: x.is_readable(), connections)
        writable = filter(lambda x: x.is_writable(), connections)
        try:
            retval = select.select(readable, writable, connections, timeout)
        except select.error, err:
            # What the heck?  Could be a window resize in curses mode..
            broadcast_event('select-error', err)
            continue
        
        ready = retval[1] # writables are more important
        for r in retval[0]: # add any readables
            if r not in ready: ready.append(r)

        if ready:
            timeout = 5.0;
            
            # Pick one socket, and prefer the highest priority handlers
            priorities = map(lambda x: x.priority, ready) # priorities of all
            max_priority = max(priorities) # highest priority
            ready = filter(lambda x, m=max_priority: x.priority==m, ready)
            broadcast_event('status', ready, connections, ready[0])
            ready = ready[0]

            i = connections.index(ready)
            try:
                # The result of processing this handler is a list
                # of replacement handlers.  In some sense this is
                # like the Actor model, where an actor processes
                # input and then replaces itself with other actors.
                # We put the new handlers at the end to give other
                # handlers a chance to run.
                del connections[i]
                before_time = time()
                connections = connections + ready.process()
                after_time = time()
                time_elapsed = after_time - before_time
                broadcast_event('time-handle', ready, time_elapsed)
            except IOError, err:
                # If there was an I/O error, let's er um, print it
                sys.last_traceback = None
                broadcast_event('io-error', ready, err)
            except:
                # If there was a socket error, print it out
                import traceback
                traceback.print_exc(100, open('/tmp/proxy.traceback', 'a'))
                traceback.print_exc()
                sys.last_traceback = None
        else:
            broadcast_event('idle', connections)
            # There's nothing to do, so increase the timeout
            if timeout < 1000.0: timeout = timeout * 2.0

            # Ping all the connections
            new_conn = []
            for c in connections:
                new_conn = new_conn + c.ping()
            connections = new_conn

            # Since there's nothing to do, let's check if we should
            # reload any modules
            # modules = [proxy3_block, proxy3_html]
            # proxy3_html isn't being imported, yuck
            # check file time, reload/recompile

def main():
    try: proxy()
    except socket.error, err:
        import errno
        if err[0] == errno.EADDRINUSE:
            print '** There is already a program running at this port. **'
        else:
            print '** Socket error:', err
    except KeyboardInterrupt:
        print 'Terminating...'
    broadcast_event('proxy-terminate')

from proxy3_util import *
from proxy3_event import broadcast_event
import proxy3_base, proxy3_options, proxy3_web, proxy3_config

if __name__ == '__main__': main()
