# This file keeps the global filter table.

import time
from string import *
from proxy3_util import *
from proxy3_base import CopyData, DataReader, Handler
from proxy3_event import broadcast_event

filter_table = []

def register_filter(pattern, mimetype, filter):
    """option is a string (read from the options structure) or None;
    pattern is a URL patter, with wildcards;
    mimetype is a MIME type, with wildcards;
    filter is the filter class to add, or None to cancel all filters"""

    pattern = glob_to_regex(pattern)
    mimetype = glob_to_regex(mimetype)
    filter_table.append( (pattern, mimetype, filter) )

def get_filter(url, type, parent, length, clientheaders, serverheaders):
    # Find the substitutions appropriate for this document
    stream = CopyData(parent, length)
    filters = [] # remember which filters we have applied
    table = filter_table[:]
    table.reverse() # Go through backwards, stopping at None

    # The mime type might have multiple parts; we only consider the first.
    if find(type, ';') >= 0: type = type[:find(type, ';')]
    
    for urlpat,mimetype,result in table:
	if urlpat.match(url) != None and mimetype.match(type) != None:
	    if result is None:
		break
	    elif result not in filters:
		stream = result(parent, stream, length, url=url,
                                type=type, clientheaders=clientheaders,
                                serverheaders=serverheaders)
		stream.socket = None
		filters.append(result)
    return stream

class Filter(DataReader):
    """Base class for all filter objects"""

    def __init__(self, parent, next, length, kw):
	DataReader.__init__(self, parent, length)
	if next: self.outfile = next
	self.args = kw

    def filter(self, s):
	"""The null filter does nothing"""
	return s

class BufferSomeFilter(Filter):
    """Save some of the data and send through filters, then write"""
    # An extension of the filter() protocol is that if filter() returns
    # two items, then the first is considered to be the string that's
    # no longer to be processed, and the second one can be re-filtered
    # at a later time.
    BUFFER_LEN = 512

    def __init__(self, parent, next, length, **kw):
	Filter.__init__(self, parent, next, length, kw)
	self.buffer = []
	self.bytesread = 0

    def read(self, data):
	self.buffer.append(data)
	self.bytesread = self.bytesread + len(data)
	if self.bytesread >= 2*self.BUFFER_LEN:
	    s = self.filter(join(self.buffer,''))
            if type(s) == type(()):
                # The first item is stuff that should be written out
                # and the second item is what should be buffered
                done, s = s
                self.outfile.write(done)
                self.bytesread = len(s)
                
	    if len(s) > self.BUFFER_LEN:
		# If there's enough data, write it to the connection
		to_write = len(s) - self.BUFFER_LEN
		self.outfile.write(s[:to_write])
		s = s[to_write:]
		self.bytesread = len(s)

	    self.buffer = [s]

    def close(self):
	Filter.close(self)
        data = self.filter(join(self.buffer,''))
        if type(data) == type(()):
            self.outfile.write(data[0])
            self.outfile.write(data[1])
        else:
            self.outfile.write(data)
	self.outfile.close()

class BufferAllFilter(Filter):
    """Save all the data and send it through a list of filters, then write"""
    
    def __init__(self, parent, next, length, **kw):
	Filter.__init__(self, parent, next, length, kw)
	self.buffer = []

    def read(self, data):
	self.buffer.append(data)

    def close(self):
	Filter.close(self)
	self.outfile.write(self.filter(join(self.buffer,'')))
	self.outfile.close()

class UnChunkingFilter(Filter):
    """Read chunked data and write out unchunked data"""

    # See HTTP/1.1 spec for the chunked transfer encoding
    def __init__(self, parent, next, length, **kw):
	Filter.__init__(self, parent, next, length, kw)
	self.chunksize = 0

    def process(self, errcount=0):
	# print 'processing unchunker', self.chunksize
	if self.chunksize <= 0:
	    # Read in a chunk size
	    self.socket.setblocking(1)
	    line = self.infile.readline()
	    self.socket.setblocking(0)
	    # print 'processing chunksize', `line`
	    while not split(line):
		line = self.infile.readline()
		# print 'rereading', `line`
	    # The first word should be a chunk size, in hex format
	    try:
		self.chunksize = atoi(split(line)[0], 16)
		# print '***CHUNKSIZE = ', self.chunksize
	    except:
		# Couldn't parse, let's give up
		self.close()
		return []

	    if self.chunksize == 0:
		# End of document, so we may need to read any extra headers
		if self.args['extra_headers']:
		    rfc822.Message(self.infile, 0) # ignore them
		self.close()
		return []

	try: 
	    data = self.infile.read(min(512, self.chunksize))
	    #print '***', len(data),'/',min(512,self.chunksize)
	    #print `data`
	except IOError, err:
	    if err.errno == 0: # See DataReader
		# print 'Unix-error'
		broadcast_event('unix-error')
		if errcount >= 5:  return [self]
		else:              return self.process(1+errcount)
	    broadcast_event('io-error', self, err)
	    data = None
	if not data:   # This connection is dead
	    self.close()
	    return []
	else:          # This connection may have more data
	    self.bytes = self.bytes + len(data)
	    self.chunksize = self.chunksize - len(data)
	    self.last_active = time.time()
	    try: self.read(data)
	    except IOError, err:   # Broken pipe?
		broadcast_event('io-error', self, err)
		self.socket.shutdown(2)
		self.close()
		return []
	    if self.length and self.bytes >= self.length: 
		self.close()
		if self.print_sizes: print '|      ', `self`
		return [] # We've read enough
	    return [self]

    def write(self, data):
	"Oddly enough, write(x) is an alias for read(x)"
	return self.read(data)

    def read(self, data):
	self.outfile.write(data)

    def close(self):
	self.infile.close()
	self.outfile.close()

import proxy3_options
